package main

import (
	"bytes"
	"encoding/binary"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net"
	"net/http"
	"os"
	"time"

	"golang.org/x/text/encoding/korean"
	"golang.org/x/text/transform"
	"gopkg.in/yaml.v2"
)

// Config 구조체 정의
type Config struct {
	ServerAddress     string `yaml:"server_address"`
	MaxRetries        int    `yaml:"max_retries"`
	ReconnectInterval int    `yaml:"reconnect_interval"`
	ReadTimeout       int    `yaml:"read_timeout"`
	RestAPIEndpoint   string `yaml:"rest_api_endpoint"`
}

// LogEntry 구조체 정의 (이전과 동일)
type LogEntry struct {
	Size     byte
	Type     uint8
	DateTime time.Time
	On       uint8
	Address  [7]byte
	Area     [24]byte
	Device   [24]byte
}

// JSONLogEntry 구조체 정의 (이전과 동일)
type JSONLogEntry struct {
	Type     uint8     `json:"type"`
	DateTime time.Time `json:"dateTime"`
	On       uint8     `json:"on"`
	Address  string    `json:"address"`
	Area     string    `json:"area"`
	Device   string    `json:"device"`
}

var config Config

func loadConfig() error {
	data, err := ioutil.ReadFile("config.yaml")
	if err != nil {
		return err
	}

	return yaml.Unmarshal(data, &config)
}

func main() {
	if err := loadConfig(); err != nil {
		fmt.Printf("Error loading config: %v\n", err)
		os.Exit(1)
	}

	for {
		if err := runClient(); err != nil {
			fmt.Printf("Error: %v\n", err)
			time.Sleep(time.Duration(config.ReconnectInterval) * time.Second)
		}
	}
}

func runClient() error {
	var conn net.Conn
	var err error

	for i := 0; i < config.MaxRetries; i++ {
		conn, err = net.Dial("tcp", config.ServerAddress)
		if err == nil {
			break
		}
		fmt.Printf("Connection attempt %d failed: %v\n", i+1, err)
		time.Sleep(time.Duration(config.ReconnectInterval) * time.Second)
	}

	if err != nil {
		return fmt.Errorf("failed to connect after %d attempts", config.MaxRetries)
	}
	defer conn.Close()

	fmt.Println("Connected to server")

	for {
		if err := handleConnection(conn); err != nil {
			return err
		}
	}
}

func handleConnection(conn net.Conn) error {
	conn.SetReadDeadline(time.Now().Add(time.Duration(config.ReadTimeout) * time.Second))

	buf := make([]byte, 64)
	_, err := conn.Read(buf)
	if err != nil {
		return fmt.Errorf("read error: %v", err)
	}

	logEntry, err := Deserialize(buf)
	if err != nil {
		return fmt.Errorf("deserialization error: %v", err)
	}

	jsonData, err := convertToJSON(logEntry)
	if err != nil {
		return fmt.Errorf("JSON conversion error: %v", err)
	}

	if err := sendToRESTAPI(jsonData); err != nil {
		return fmt.Errorf("REST API error: %v", err)
	}

	return nil
}

func Deserialize(data []byte) (*LogEntry, error) {
	if len(data) != 64 {
		return nil, fmt.Errorf("invalid data length: expected 64, got %d", len(data))
	}

	le := &LogEntry{}
	le.Size = data[0]
	le.Type = data[1]
	year := binary.LittleEndian.Uint16(data[2:4])
	le.DateTime = time.Date(int(year), time.Month(data[4]), int(data[5]), int(data[6]), int(data[7]), int(data[8]), 0, time.UTC)
	le.On = data[9]
	copy(le.Address[:], data[10:17])
	copy(le.Area[:], data[17:41])
	copy(le.Device[:], data[41:])

	return le, nil
}

func convertToJSON(le *LogEntry) ([]byte, error) {
	area, err := decodeEUCKR(bytes.TrimRight(le.Area[:], "\x00"))
	if err != nil {
		return nil, err
	}

	device, err := decodeEUCKR(bytes.TrimRight(le.Device[:], "\x00"))
	if err != nil {
		return nil, err
	}

	jsonLE := JSONLogEntry{
		Type:     le.Type,
		DateTime: le.DateTime,
		On:       le.On,
		Address:  string(bytes.TrimRight(le.Address[:], "\x00")),
		Area:     area,
		Device:   device,
	}

	return json.Marshal(jsonLE)
}

func decodeEUCKR(b []byte) (string, error) {
	decoder := korean.EUCKR.NewDecoder()
	result, _, err := transform.Bytes(decoder, b)
	if err != nil {
		return "", err
	}
	return string(result), nil
}

func sendToRESTAPI(data []byte) error {
	resp, err := http.Post(config.RestAPIEndpoint, "application/json", bytes.NewBuffer(data))
	if err != nil {
		return err
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		return fmt.Errorf("unexpected status code: %d", resp.StatusCode)
	}

	return nil
}
